#include "../includeCXX/sfx_surface.h"
#include <cmath>
#include <iostream>

template<typename T>
void get_charnock_roughness(T &z0_m, T &u_dyn0,
    const T h, const T U,
    const T kappa, 
    const T h_charnock, const T c1_charnock, const T c2_charnock, 
    const int maxiters)
{
    T Uc, a, b, c, c_min, f;

    Uc = U;
    a = 0.0;
    b = 25.0;
    c_min = log(h_charnock) / kappa;

    for (int i = 0; i < maxiters; i++)
    {
        f = c1_charnock - 2.0 * log(Uc);
        for (int j = 0; j < maxiters; j++)
        {
            c = (f + 2.0 * log(b)) / kappa;
            if (U <= 8.0e0) 
                a = log(1.0 + c2_charnock * ( pow(b / Uc, 3) ) ) / kappa;
            c = std::max(c - a, c_min);
            b = c;
        }
        z0_m = h_charnock * exp(-c * kappa);
        z0_m = std::max(z0_m, T(0.000015e0));
        Uc = U * log(h_charnock / z0_m) / log(h / z0_m);
    }
    
    u_dyn0 = Uc / c;
}

template void get_charnock_roughness(float &z0_m, float &u_dyn0, 
    const float h, const float U,
    const float kappa, 
    const float h_charnock, const float c1_charnock, const float c2_charnock,
    const int maxiters);
template void get_charnock_roughness(double &z0_m, double &u_dyn0,
    const double h, const double U,
    const double kappa, 
    const double h_charnock, const double c1_charnock, const double c2_charnock, 
    const int maxiters);

template<typename T>
void get_thermal_roughness(T &z0_t, T &B,
    const T z0_m, const T Re, 
    const T Re_rough_min, 
    const T B1_rough, const T B2_rough, const T B3_rough, const T B4_rough, 
    const T B_max_ocean, const T B_max_lake, const T B_max_land,
    const int surface_type)
{
    // --- define B = log(z0_m / z0_t)
    if (Re <= Re_rough_min) 
        B = B1_rough * log(B3_rough * Re) + B2_rough;
    else
        // *: B4 takes into account Re value at z' ~ O(10) z0
        B = B4_rough * (pow(Re, B2_rough));

    //   --- apply max restriction based on surface type
    if (surface_type == 0) 
        B = std::min(B, B_max_ocean);
    else if (surface_type == 1) 
        B = std::min(B, B_max_lake);
    else if (surface_type == 2)
        B = std::min(B, B_max_land);

    // --- define roughness [thermal]
    z0_t = z0_m / exp(B);
}

template void get_thermal_roughness(float &z0_t, float &B,
    const float z0_m, const float Re, 
    const float Re_rough_min, 
    const float B1_rough, const float B2_rough, const float B3_rough, const float B4_rough, 
    const float B_max_ocean, const float B_max_lake, const float B_max_land,
    const int surface_type);
template void get_thermal_roughness(double &z0_t, double &B,
    const double z0_m, const double Re, 
    const double Re_rough_min, 
    const double B1_rough, const double B2_rough, const double B3_rough, const double B4_rough, 
    const double B_max_ocean, const double B_max_lake, const double B_max_land,
    const int surface_type);